from pyspark.sql.dataframe import DataFrame
from asyncactions.utils import patch_all


actions = {
    "collect": ("""
        Returns a `concurrent.futures.Future` for retrieving all elements of this DataFrame.

        >>> df = spark.range(8)             # doctest: +SKIP
        >>> f = df.collectAsync()            # doctest: +SKIP
        >>> f.result()                       # doctest: +SKIP
        [Row(id=0), Row(id=1), Row(id=2), Row(id=3), Row(id=4), Row(id=5), Row(id=6), Row(id=7)]
        """),
    "count": ("""
        >>> df = spark.range(10)             # doctest: +SKIP
        >>> f = df.countAsync()              # doctest: +SKIP
        >>> f.result()                       # doctest: +SKIP
        10
        """),
    "foreach": (""""
        Asynchronously applies a function f to all elements of this DataFrame
        and returns a `concurrent.futures.Future` of this action.

        >>> def g(x): print(x)               # doctest: +SKIP
        >>> df = spark.range(10)             # doctest: +SKIP
        >>> f = df.foreachAsync(g)           # doctest: +SKIP
        >>> f.result() is None               # doctest: +SKIP
        True
        """),
    "foreachPartition": ("""
        Asynchronously applies a function f to each partition of this DataFrame
        and returns a `concurrent.futures.Future` of this action.

        >>> def g(xs):                       # doctest: +SKIP
        ...     for x in xs:
        ...         print(x)
        >>> df = spark.range(10)             # doctest: +SKIP
        >>> f = df.foreachPartitionAsync(g)  # doctest: +SKIP
        >>> f.result() is None               # doctest: +SKIP
        """),
    "take": ("""
        Returns a `concurrent.futures.Future` for retrieving
        the first num elements of the DataFrame.

        >>> rdd = spark.range(10)            # doctest: +SKIP
        >>> f = df.takeAsync(3)              # doctest: +SKIP
        >>> f.result() is None               # doctest: +SKIP
        [Row(id=0), Row(id=1), Row(id=2)]
        """)
}


patch_all(DataFrame, actions)



from pyspark.rdd import RDD
from asyncactions.utils import patch_all


actions = {
    "collect": ("""
        Returns a `concurrent.futures.Future` for retrieving all elements of this RDD.

        >>> rdd = sc.parallelize(range(10))  # doctest: +SKIP
        >>> f = rdd.collectAsync()           # doctest: +SKIP
        >>> f.result()                       # doctest: +SKIP
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        """),
    "count": ("""
        Return `concurrent.futures.Future` of the number of elements in this RDD.

        >>> rdd = sc.parallelize(range(10))  # doctest: +SKIP
        >>> f = rdd.countAsync()             # doctest: +SKIP
        >>> f.result()                       # doctest: +SKIP
        10
        """),
    "foreach": (""""
        Asynchronously applies a function f to all elements of this RDD
        and returns a `concurrent.futures.Future` of this action.

        >>> def g(x): print(x)               # doctest: +SKIP
        >>> rdd = sc.parallelize(range(10))  # doctest: +SKIP
        >>> f = rdd.foreachAsync(g)          # doctest: +SKIP
        >>> f.result() is None               # doctest: +SKIP
        True
        """),
    "foreachPartition": ("""
        Asynchronously applies a function f to each partition of this RDD
        and returns a `concurrent.futures.Future` of this action.

        >>> def g(xs):                       # doctest: +SKIP
        ...     for x in xs:
        ...         print(x)
        >>> rdd = sc.parallelize(range(10))  # doctest: +SKIP
        >>> f = rdd.foreachPartitionAsync(g) # doctest: +SKIP
        >>> f.result() is None               # doctest: +SKIP
        """),
    "take": ("""
        Returns a `concurrent.futures.Future` for retrieving
        the first num elements of the RDD.

        >>> rdd = sc.parallelize(range(10))  # doctest: +SKIP
        >>> f = rdd.takeAsync(3)             # doctest: +SKIP
        >>> f.result() is None               # doctest: +SKIP
        [0, 1, 2]
        """),
    "saveAsTextFile": ("""
        Asynchronously save this RDD as a text file, using string representations of elements
        and returns `concurrent.futures.Future` of this action.
        
        @param path: path to text file
        @param compressionCodecClass: (None by default) string i.e.
            "org.apache.hadoop.io.compress.GzipCodec"
    """)
}

patch_all(RDD, actions)



import os
import shutil
import tempfile
import unittest

from pyspark.tests import ReusedPySparkTestCase
import asyncactions


class AsyncRDDActionsTestCase(ReusedPySparkTestCase):
    @classmethod
    def setUpClass(cls):
        ReusedPySparkTestCase.setUpClass()
        cls.tempdir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        ReusedPySparkTestCase.tearDownClass()
        shutil.rmtree(cls.tempdir)

    def test_async_actions(self):
        data = [x for x in range(10)]
        rdd = self.sc.parallelize(data)
        f = rdd.collectAsync()
        self.assertListEqual(f.result(), data)

        f = rdd.countAsync()
        self.assertEqual(f.result(), 10)

        f = rdd.takeAsync(5)
        self.assertEqual(f.result(), data[:5])

        acc1 = self.sc.accumulator(0)

        f = rdd.foreachAsync(lambda _: acc1.add(1))
        self.assertTrue(
            f.result() is None and acc1.value == len(data)
        )

        acc2 = self.sc.accumulator(0)
        f = rdd.foreachPartitionAsync(lambda xs: [acc2.add(1) for _ in xs])
        self.assertTrue(
            f.result() is None and acc2.value == len(data)
        )

        path = os.path.join(self.tempdir, "rdd_saved_async")
        f = rdd.saveAsTextFileAsync(path)

        self.assertIsNone(f.result())

        self.assertEqual(
            sorted(self.sc.textFile(path).collect()),
            [str(x) for x in data]
        )


if __name__ == '__main__':
    unittest.main()



from pyspark.sql.readwriter import DataFrameWriter
from asyncactions.utils import patch_all


actions = {
    "save": """
        Asynchronously saves the contents of the :class:`DataFrame` to a data source
        and returns a `concurrent.futures.Future` of this action.

        The data source is specified by the ``format`` and a set of ``options``.
        If ``format`` is not specified, the default data source configured by
        ``spark.sql.sources.default`` will be used.

        :param path: the path in a Hadoop supported file system
        :param format: the format used to save
        :param mode: specifies the behavior of the save operation when data already exists.

        * ``append``: Append contents of this :class:`DataFrame` to existing data.
        * ``overwrite``: Overwrite existing data.
        * ``ignore``: Silently ignore this operation if data already exists.
        * ``error`` (default case): Throw an exception if data already exists.
        :param partitionBy: names of partitioning columns
        :param options: all other string options

        >>> _ = (df.write.mode('append').format("parquet")              # doctest: +SKIP
        ...     .saveAsync(os.path.join(tempfile.mkdtemp(), 'data')))   
        """
}


patch_all(DataFrameWriter, actions)



import concurrent.futures
from pyspark.context import SparkContext


def _get_executor(self):
    """ Return existing thread pool executor
    or create a new one.
    """
    # This would fail anyway, but
    # we don't want an orphan executor
    if SparkContext._active_spark_context is None:
        raise ValueError("No active SparkContext")

    if getattr(SparkContext, "_thread_pool_executor", None) is None:
        try:
            import concurrent.futures

            # Make sure that there is only one executor
            with SparkContext._lock:
                cores = self.getConf().get("spark.driver.cores") or 2
                # In case of another thread got lock first
                SparkContext._thread_pool_executor = (
                    getattr(SparkContext, "_thread_pool_executor", None) or
                    concurrent.futures.ThreadPoolExecutor(max_workers=cores)
                )

        # Python 2.7 and not futures backport installed
        except ImportError as e:
            msg = "{}. Async actions require Python >= 3.2 or futures package installed"
            raise ImportError(msg.format(e.message))

    return SparkContext._thread_pool_executor


def stop(self):
    with SparkContext._lock:
        if getattr(SparkContext, "_thread_pool_executor", None):
            SparkContext._thread_pool_executor.shutdown()
            SparkContext._thread_pool_executor = None
    SparkContext._stop(self)


SparkContext._stop = SparkContext.stop
SparkContext.stop = stop
SparkContext._get_executor = _get_executor



import sys
import os
from setuptools import setup, find_packages


install_requires = [] if sys.version_info >= (3, 2) else ["futures"]

setup(
    name='pyspark-asyncactions',
    version='0.0.1',
    packages=find_packages("."),
    url='https://github.com/zero323/pyspark-asyncactions',
    license='Apache 2.0',
    author='zero323',
    author_email='',
    description='A proof of concept asynchronous actions for PySpark using concurent.futures',
    long_description=(open('README.rst').read() if os.path.exists('README.rst')
                        else ''), 
    install_requires=install_requires,
    classifiers=['Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 3',
                 'Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: Apache Software License',
                ]
)



import tempfile
import shutil
import os
import unittest

from pyspark.tests import ReusedPySparkTestCase
from pyspark.sql import SparkSession
from pyspark.sql.types import Row

import asyncactions


class AsyncDataFrameActionsTestCase(ReusedPySparkTestCase):
    @classmethod
    def setUpClass(cls):
        ReusedPySparkTestCase.setUpClass()
        cls.spark = SparkSession(cls.sc)

    @classmethod
    def tearDownClass(cls):
        ReusedPySparkTestCase.tearDownClass()
        cls.spark.stop()

    def test_async_actions(self):
        data = [Row(id=i) for i in range(10)]
        df = self.spark.createDataFrame(data)
        f = df.collectAsync()
        self.assertListEqual(f.result(), data)

        f = df.countAsync()
        self.assertEqual(f.result(), 10)

        f = df.takeAsync(3)
        self.assertEqual(f.result(), data[:3])

        acc1 = self.sc.accumulator(0)

        f = df.foreachAsync(lambda _: acc1.add(1))
        self.assertTrue(
            f.result() is None and acc1.value == len(data)
        )

        acc2 = self.sc.accumulator(0)
        f = df.foreachPartitionAsync(lambda xs: [acc2.add(1) for _ in xs])
        self.assertTrue(
            f.result() is None and acc2.value == len(data)
        )


class AsyncDataFrameWriterActionsTestCase(ReusedPySparkTestCase):
    @classmethod
    def setUpClass(cls):
        ReusedPySparkTestCase.setUpClass()
        cls.spark = SparkSession(cls.sc)
        cls.tempdir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        ReusedPySparkTestCase.tearDownClass()
        cls.spark.stop()
        shutil.rmtree(cls.tempdir)

    def test_save_async(self):
        data = [Row(id=i) for i in range(10)]
        df = self.spark.createDataFrame(data)

        path = os.path.join(self.tempdir, "saved_async")
        f = (df
             .write
             .format("json")
             .saveAsync(path))

        self.assertIsNone(f.result())
        loaded = self.spark.read.format("json").load(path)
        self.assertEqual(loaded.count(), 10)
        self.assertEqual(sorted(loaded.collect()), data)


if __name__ == '__main__':
    unittest.main()



import textwrap
from functools import update_wrapper

from pyspark.context import SparkContext
from pyspark.rdd import RDD
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.readwriter import DataFrameWriter


def get_context(x):
    if isinstance(x, RDD):
        return x.ctx
    if isinstance(x, DataFrame):
        return x._sc
    if isinstance(x, DataFrameWriter):
        return x._spark._sc


def async_action(f):
    def async_action_(self, *args, **kwargs):
        executor = get_context(self)._get_executor()
        return executor.submit(f, self, *args, **kwargs)
    return async_action_


def patch_async(cls, method, doc, suffix):
    g = getattr(cls, method)
    f = update_wrapper(async_action(g), g, assigned=["__module__", "__annotations__"] if hasattr(g, "__annotations__") else ["__module__"])
    f.__doc__ = textwrap.dedent(doc)
    setattr(cls, "{}{}".format(method, suffix), f)


def patch_all(cls, mapping, suffix="Async"):
    for method, doc in mapping.items():
        patch_async(cls, method, doc, suffix)
